<?php

declare(strict_types=1);

namespace Drupal\otlrs\EventSubscriber;

use Drupal\otlrs\HttpServerSpanAttributesTrait;
use OpenTelemetry\API\Trace\LocalRootSpan;
use OpenTelemetry\API\Trace\SpanInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Event\ExceptionEvent;
use Symfony\Component\HttpKernel\Event\RequestEvent;
use Symfony\Component\HttpKernel\KernelEvents;

/**
 * Add HTTP route and exception attributes to the local root span.
 *
 * @see https://opentelemetry.io/docs/specs/semconv/http/http-spans/#http-server-semantic-conventions
 */
final class LocalRootSpanSubscriber implements EventSubscriberInterface {

  use HttpServerSpanAttributesTrait;

  /**
   * Constructs local root span subscriber.
   *
   * @param \OpenTelemetry\API\Trace\SpanInterface $span
   *   OpenTelemetry span, only use in unit tests.
   */
  public function __construct(
    private ?SpanInterface $span = NULL,
  ) {}

  /**
   * Kernel request event handler.
   *
   * @param \Symfony\Component\HttpKernel\Event\RequestEvent $event
   *   The request event.
   */
  public function onKernelRequest(RequestEvent $event): void {
    if ($event->isMainRequest()) {
      $this->updateSpan($this->currentSpan(), $event->getRequest());
    }
  }

  /**
   * Kernel exception event handler.
   *
   * @param \Symfony\Component\HttpKernel\Event\ExceptionEvent $event
   *   The exception event.
   */
  public function onKernelException(ExceptionEvent $event): void {
    if ($event->isMainRequest()) {
      $span = $this->currentSpan();
      $span->recordException($event->getThrowable());
      $this->updateSpan($span, $event->getRequest());
    }
  }

  /**
   * Update the span with request method and target (route name).
   *
   * @param \OpenTelemetry\API\Trace\SpanInterface $span
   *   The OpenTelemetry span.
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The request object.
   */
  private function updateSpan(SpanInterface $span, Request $request): void {
    $routeAttributes = $this->getRouteAttributes($request);
    if (!empty($routeAttributes)) {
      $span->setAttributes($routeAttributes);

      // Update span name to include target route.
      $methodAttributes = $this->getRequestMethodAttributes($request);
      $nameAttributes = [...$routeAttributes, ...$methodAttributes];
      $span->updateName($this->getSpanName($nameAttributes));
    }
  }

  /**
   * Returns the current local root span.
   */
  private function currentSpan(): SpanInterface {
    return $this->span ?? LocalRootSpan::current();
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents(): array {
    return [
      // Just after \Symfony\Component\HttpKernel\EventListener\RouterListener.
      KernelEvents::REQUEST => [['onKernelRequest', 31]],
      // Same as \Drupal\Core\EventSubscriber\ExceptionLoggingSubscriber.
      KernelEvents::EXCEPTION => [['onKernelException', 50]],
    ];
  }

}
