<?php

declare(strict_types=1);

namespace Drupal\otlrs;

use Drupal\Core\Routing\RouteMatch;
use OpenTelemetry\SDK\Common\Configuration\Configuration;
use OpenTelemetry\SemConv\TraceAttributes;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;

/**
 * Provides methods to gather HTTP server span attributes.
 *
 * @see https://opentelemetry.io/docs/specs/semconv/http/http-spans/#http-server-semantic-conventions
 */
trait HttpServerSpanAttributesTrait {

  /**
   * Setting key to override the list of known request methods.
   */
  private const KNOWN_METHODS_KEY = 'OTEL_INSTRUMENTATION_HTTP_KNOWN_METHODS';

  /**
   * Setting key to specify the list of request headers to record.
   */
  private const REQUEST_HEADERS_KEY = 'OTEL_INSTRUMENTATION_HTTP_CAPTURE_HEADERS_SERVER_REQUEST';

  /**
   * Setting key to specify the list of response headers to record.
   */
  private const RESPONSE_HEADERS_KEY = 'OTEL_INSTRUMENTATION_HTTP_CAPTURE_HEADERS_SERVER_RESPONSE';

  /**
   * Default list of known request methods.
   */
  private const KNOWN_METHODS = [
    'GET',
    'HEAD',
    'POST',
    'PUT',
    'DELETE',
    'CONNECT',
    'OPTIONS',
    'TRACE',
    'PATCH',
  ];

  /**
   * Returns the span name determined from span attributes.
   *
   * HTTP span names SHOULD be {method} {target} if there is a (low-cardinality)
   * target available. If there is no (low-cardinality) {target} available, HTTP
   * span names SHOULD be {method}.
   *
   * The {method} MUST be {http.request.method} if the method represents the
   * original method known to the instrumentation. In other cases (when
   * {http.request.method} is set to _OTHER), {method} MUST be HTTP.
   *
   * The {target} SHOULD be http.route for HTTP Server spans. Instrumentation
   * MUST NOT default to using URI path as a {target}.
   *
   * @param array<string, mixed> $attributes
   *   Span attributes as key-value pairs.
   *
   * @return string
   *   The span name determined from request method and route (if any).
   *
   * @see https://opentelemetry.io/docs/specs/semconv/http/http-spans/#name
   */
  protected function getSpanName(array $attributes): string {
    $result = 'HTTP';

    if (
      isset($attributes[TraceAttributes::HTTP_REQUEST_METHOD]) &&
      $attributes[TraceAttributes::HTTP_REQUEST_METHOD] !== '_OTHER'
    ) {
      $result = $attributes[TraceAttributes::HTTP_REQUEST_METHOD];
    }
    if (isset($attributes[TraceAttributes::HTTP_ROUTE])) {
      $result .= " {$attributes[TraceAttributes::HTTP_ROUTE]}";
    }

    return $result;
  }

  /**
   * Returns request method attributes.
   *
   * HTTP request method value SHOULD be "known" to the instrumentation. By
   * default, this convention defines "known" methods as the ones listed in
   * RFC9110 and the PATCH method defined in RFC5789.
   *
   * If the HTTP request method is not known to instrumentation, it MUST set the
   * http.request.method attribute to _OTHER.
   *
   * If the HTTP instrumentation could end up converting valid HTTP request
   * methods to _OTHER, then it MUST provide a way to override the list of known
   * HTTP methods. If this override is done via environment variable, then the
   * environment variable MUST be named OTEL_INSTRUMENTATION_HTTP_KNOWN_METHODS
   * and support a comma-separated list of case-sensitive known HTTP methods
   * (this list MUST be a full override of the default known method, it is not a
   * list of known methods in addition to the defaults).
   *
   * HTTP method names are case-sensitive and http.request.method attribute
   * value MUST match a known HTTP method name exactly. Instrumentations for
   * specific web frameworks that consider HTTP methods to be case insensitive,
   * SHOULD populate a canonical equivalent. Tracing instrumentations that do
   * so, MUST also set http.request.method_original to the original value.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The request object.
   *
   * @return array<string, string>
   *   Span attributes as key-value pairs.
   */
  protected function getRequestMethodAttributes(Request $request): array {
    $attributes = [];

    $method = $request->getMethod();
    $methods = $this->getConfigurationList(static::KNOWN_METHODS_KEY, static::KNOWN_METHODS);
    $parsedMethod = in_array($method, $methods, strict: TRUE) ? $method : '_OTHER';
    $attributes[TraceAttributes::HTTP_REQUEST_METHOD] = $parsedMethod;

    $originalMethod = $request->server->get('REQUEST_METHOD', 'GET');
    if ($parsedMethod !== $originalMethod) {
      $attributes[TraceAttributes::HTTP_REQUEST_METHOD_ORIGINAL] = $originalMethod;
    }

    return $attributes;
  }

  /**
   * Returns request protocol attributes.
   *
   * Attribute network.control.version: The actual version of the protocol used
   * for network communication.
   *
   * Attribute network.protocol.name: SHOULD be normalized to lowercase. If not
   * http and network.protocol.version is set.
   *
   * @return array<string, string>
   *   Span attributes as key-value pairs.
   */
  protected function getRequestProtocolAttributes(Request $request): array {
    $attributes = [];

    [$name, $version] = explode('/', strtolower($request->getProtocolVersion()));
    $attributes[TraceAttributes::NETWORK_PROTOCOL_VERSION] = $version;
    if ($name !== 'http') {
      $attributes[TraceAttributes::NETWORK_PROTOCOL_NAME] = $name;
    }

    return $attributes;
  }

  /**
   * Returns the HTTP route attribute if available.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The request object.
   *
   * @return array<string, string>
   *   Span attributes as key-value pairs.
   */
  protected function getRouteAttributes(Request $request): array {
    $attributes = [];

    $routeObject = RouteMatch::createFromRequest($request)->getRouteObject();
    if ($routeObject !== NULL) {
      $attributes[TraceAttributes::HTTP_ROUTE] = $routeObject->getPath();
    }

    return $attributes;
  }

  /**
   * Returns a list of request header attributes to include.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The request object.
   *
   * @return array<string, string>
   *   Span attributes as key-value pairs.
   */
  protected function getRequestHeaderAttributes(Request $request): array {
    $attributes = [];
    foreach ($this->getConfigurationList(static::REQUEST_HEADERS_KEY) as $name) {
      if ($request->headers->has($name)) {
        $key = TraceAttributes::HTTP_REQUEST_HEADER . '.' . strtolower($name);
        $attributes[$key] = $request->headers->all($name);
      }
    }
    return $attributes;
  }

  /**
   * Returns a list of response header attributes to include.
   *
   * @param \Symfony\Component\HttpFoundation\Response $response
   *   The request object.
   *
   * @return array<string, string>
   *   Span attributes as key-value pairs.
   */
  protected function getResponseHeaderAttributes(Response $response): array {
    $attributes = [];
    foreach ($this->getConfigurationList(static::RESPONSE_HEADERS_KEY) as $name) {
      if ($response->headers->has($name)) {
        $key = TraceAttributes::HTTP_RESPONSE_HEADER . '.' . strtolower($name);
        $attributes[$key] = $response->headers->all($name);
      }
    }
    return $attributes;
  }

  /**
   * Returns list of config values parsed from environment or SDK configuration.
   *
   * @param string $key
   *   The config key.
   * @param ?array $default
   *   A default value returned when no config with the given key exists.
   *
   * @return string[]
   *   A list of strings parsed from the config.
   */
  protected function getConfigurationList(string $key, ?array $default = []): array {
    $env = getenv($key);
    return match (TRUE) {
      is_string($env) => array_map(trim(...), explode(',', $env)),
      class_exists(Configuration::class) && Configuration::has($key) => Configuration::getList($key),
      default => $default
    };
  }

}
