<?php

declare(strict_types=1);

namespace Drupal\otlrs\StackMiddleware;

use Drupal\otlrs\HttpServerSpanAttributesTrait;
use OpenTelemetry\API\Trace\LocalRootSpan;
use OpenTelemetry\API\Trace\SpanInterface;
use OpenTelemetry\SemConv\TraceAttributes;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\HttpKernelInterface;

/**
 * Add attributes based on request / response headers to the local root span.
 *
 * @see https://opentelemetry.io/docs/specs/semconv/http/http-spans/#http-server-semantic-conventions
 */
final class LocalRootSpanMiddleware implements HttpKernelInterface {

  use HttpServerSpanAttributesTrait;

  /**
   * Constructs an OtAttrsResponseMiddleware object.
   *
   * @param \Symfony\Component\HttpKernel\HttpKernelInterface $httpKernel
   *   Next http kernel in the stack.
   * @param \OpenTelemetry\API\Trace\SpanInterface $span
   *   OpenTelemetry span, only use in unit tests.
   */
  public function __construct(
    private readonly HttpKernelInterface $httpKernel,
    private ?SpanInterface $span = NULL,
  ) {}

  /**
   * {@inheritdoc}
   */
  public function handle(Request $request, $type = self::MAIN_REQUEST, $catch = TRUE): Response {
    if ($type === self::MAIN_REQUEST) {
      $this->updateSpanFromRequest($this->currentSpan(), $request);
    }

    $response = $this->httpKernel->handle($request, $type, $catch);

    if ($type === self::MAIN_REQUEST) {
      $span = $this->currentSpan();

      // Required attributes.
      $span->setAttribute(TraceAttributes::HTTP_RESPONSE_STATUS_CODE, $response->getStatusCode());

      // Opt-In attributes.
      $span->setAttributes($this->getResponseHeaderAttributes($response));
    }

    return $response;
  }

  /**
   * Update the span with attributes from the incoming request.
   *
   * @param \OpenTelemetry\API\Trace\SpanInterface $span
   *   The OpenTelemetry span.
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The request object.
   */
  private function updateSpanFromRequest(SpanInterface $span, Request $request): void {
    // Required attributes.
    $attributes = $this->getRequestMethodAttributes($request);
    $span->setAttributes($attributes);
    $span->updateName($this->getSpanName($attributes));
    $span->setAttribute(TraceAttributes::URL_PATH, $request->getPathInfo());
    $span->setAttribute(TraceAttributes::URL_SCHEME, $request->getScheme());
    if ($request->query->count() > 0) {
      $span->setAttribute(TraceAttributes::URL_QUERY, $request->getQueryString());
    }

    // Recommended attributes.
    $span->setAttribute(TraceAttributes::CLIENT_ADDRESS, $request->getClientIp());
    $span->setAttribute(TraceAttributes::NETWORK_PEER_ADDRESS, $request->server->get('REMOTE_ADDR'));
    $span->setAttribute(TraceAttributes::SERVER_ADDRESS, $request->getHost());
    $span->setAttribute(TraceAttributes::SERVER_PORT, $request->getPort());
    $span->setAttribute(TraceAttributes::USER_AGENT_ORIGINAL, $request->headers->get('user-agent'));
    $span->setAttributes($this->getRequestProtocolAttributes($request));

    // Opt-In attributes.
    $span->setAttributes($this->getRequestHeaderAttributes($request));
  }

  /**
   * Returns the current local root span.
   */
  private function currentSpan(): SpanInterface {
    return $this->span ?? LocalRootSpan::current();
  }

}
