<?php

declare(strict_types=1);

namespace Drupal\Tests\otlrs\Functional;

use Drupal\Component\OTUnit\FileExporterTrait;
use Drupal\Component\OTUnit\ProtoHelpersTrait;
use Drupal\Component\OTUnit\SdkAutoloadingTrait;
use Drupal\otlrs\EventSubscriber\LocalRootSpanSubscriber;
use Drupal\otlrs\StackMiddleware\LocalRootSpanMiddleware;
use Drupal\Tests\BrowserTestBase;
use Opentelemetry\Proto\Trace\V1\Span;
use Opentelemetry\Proto\Trace\V1\Span\Event;
use OpenTelemetry\SDK\Common\Configuration\Variables;
use OpenTelemetry\SemConv\TraceAttributes;
use PHPUnit\Framework\Attributes\CoversClass;
use PHPUnit\Framework\Attributes\Group;

/**
 * End-to-end test for OpenTelemetry SDK autoloading and configuration.
 */
#[Group('otlrs')]
#[CoversClass(LocalRootSpanMiddleware::class)]
#[CoversClass(LocalRootSpanSubscriber::class)]
final class LocalRootSpanIntegrationTest extends BrowserTestBase {

  use FileExporterTrait;
  use ProtoHelpersTrait;
  use SdkAutoloadingTrait;

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * {@inheritdoc}
   */
  protected static $modules = ['otlrs', 'otlrs_test'];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->setupSdkAutoloading($this->getFileExporterConfigurationVariables() + [
      Variables::OTEL_PHP_EXPERIMENTAL_AUTO_ROOT_SPAN => "true",
    ]);
  }

  /**
   * Tests that a successful page request is recorded properly.
   */
  public function testPageRequest(): void {
    $this->resetFileExporterResultFiles();
    $this->drupalGet('/otlrs_test/ok/test-param', options: ['query' => ['q' => 'foo']]);

    [$traceMessage] = $this->waitForTraces();

    // Check resource attribute.
    $traceResource = $this->getTraceResource($traceMessage);
    $sdkName = $this->getAttributeValue(
      TraceAttributes::TELEMETRY_SDK_NAME,
      $traceResource->getAttributes()
    );
    $this->assertNotNull($sdkName);
    $this->assertTrue($sdkName->hasStringValue());
    $this->assertEquals('opentelemetry', $sdkName->getStringValue());

    // Check scope.
    $scope = $this->getTraceScope($traceMessage);
    $this->assertEquals(
      'io.opentelemetry.php.auto-root-span',
      $scope->getName()
    );

    $spans = $this->getTraceSpans($traceMessage);
    assert($spans[0] instanceof Span);
    $this->assertEquals('GET /otlrs_test/ok/{param}', $spans[0]->getName());

    // Check for query span attribute, added by LocalRootSpanMiddleware.
    $urlQuery = $this->getAttributeValue(
      TraceAttributes::URL_QUERY,
      $spans[0]->getAttributes(),
    );
    $this->assertNotNull($urlQuery);
    $this->assertTrue($urlQuery->hasStringValue());
    $this->assertEquals('q=foo', $urlQuery->getStringValue());

    // Check for route span attribute, added by LocalRootSpanSubscriber.
    $httpStatus = $this->getAttributeValue(
      TraceAttributes::HTTP_ROUTE,
      $spans[0]->getAttributes()
    );
    $this->assertNotNull($httpStatus);
    $this->assertTrue($httpStatus->hasStringValue());
    $this->assertEquals('/otlrs_test/ok/{param}', $httpStatus->getStringValue());

    // Check for http response status code, added by LocalRootSpanMiddleware.
    $httpStatus = $this->getAttributeValue(
      TraceAttributes::HTTP_RESPONSE_STATUS_CODE,
      $spans[0]->getAttributes()
    );
    $this->assertNotNull($httpStatus);
    $this->assertTrue($httpStatus->hasIntValue());
    $this->assertEquals(200, $httpStatus->getIntValue());

    // Check that no exception span events were recorded.
    $events = $spans[0]->getEvents();
    $this->assertCount(0, $events);
  }

  /**
   * Tests that an exception event is recorded properly.
   */
  public function testException(): void {
    $this->resetFileExporterResultFiles();
    $this->drupalGet('/otlrs_test/error/test-param');

    [$traceMessage] = $this->waitForTraces();

    $spans = $this->getTraceSpans($traceMessage);
    assert($spans[0] instanceof Span);
    $this->assertEquals('GET /otlrs_test/error/{param}', $spans[0]->getName());

    // Check for path span attribute, added by LocalRootSpanMiddleware.
    $urlPath = $this->getAttributeValue(
      TraceAttributes::URL_PATH,
      $spans[0]->getAttributes(),
    );
    $this->assertNotNull($urlPath);
    $this->assertTrue($urlPath->hasStringValue());
    $this->assertEquals('/otlrs_test/error/test-param', $urlPath->getStringValue());

    // Check for route span attribute, added by LocalRootSpanSubscriber.
    $httpStatus = $this->getAttributeValue(
      TraceAttributes::HTTP_ROUTE,
      $spans[0]->getAttributes()
    );
    $this->assertNotNull($httpStatus);
    $this->assertTrue($httpStatus->hasStringValue());
    $this->assertEquals('/otlrs_test/error/{param}', $httpStatus->getStringValue());

    // Check for http response status code, added by LocalRootSpanMiddleware.
    $httpStatus = $this->getAttributeValue(
      TraceAttributes::HTTP_RESPONSE_STATUS_CODE,
      $spans[0]->getAttributes()
    );
    $this->assertNotNull($httpStatus);
    $this->assertTrue($httpStatus->hasIntValue());
    $this->assertEquals(500, $httpStatus->getIntValue());

    // Check that no exception span events were recorded.
    $events = $spans[0]->getEvents();
    $this->assertCount(1, $events);
    assert($events[0] instanceof Event);

    $this->assertEquals('exception', $events[0]->getName());

    // Check that exception message span attribute was recorded by
    // LocalRootSpanSubscriber.
    $exceptionMessage = $this->getAttributeValue(
      'exception.message',
      $events[0]->getAttributes()
    );
    $this->assertNotNull($exceptionMessage);
    $this->assertTrue($exceptionMessage->hasStringValue());
    $this->assertEquals('test exception', $exceptionMessage->getStringValue());

    // Check that exception stacktrace span attribute was recorded by
    // LocalRootSpanSubscriber.
    $exceptionStacktrace = $this->getAttributeValue(
      'exception.stacktrace',
      $events[0]->getAttributes()
    );
    $this->assertNotNull($exceptionStacktrace);
    $this->assertTrue($exceptionStacktrace->hasStringValue());
    $this->assertStringContainsString('TestControllers.exception', $exceptionStacktrace->getStringValue());
    $this->assertStringContainsString('LocalRootSpanMiddleware.handle', $exceptionStacktrace->getStringValue());

    // Check that exception type span attribute was recorded by
    // LocalRootSpanSubscriber.
    $exceptionType = $this->getAttributeValue(
      'exception.type',
      $events[0]->getAttributes()
    );
    $this->assertNotNull($exceptionType);
    $this->assertTrue($exceptionMessage->hasStringValue());
    $this->assertEquals('RuntimeException', $exceptionType->getStringValue());
  }

  /**
   * Tests that a denied page request is recorded properly.
   */
  public function testAccessDenied(): void {
    $this->resetFileExporterResultFiles();
    $this->drupalGet('/otlrs_test/ok/test-param/edit');

    [$traceMessage] = $this->waitForTraces();

    $spans = $this->getTraceSpans($traceMessage);
    assert($spans[0] instanceof Span);
    $this->assertEquals('GET /otlrs_test/ok/{param}/edit', $spans[0]->getName());

    // Check for query span attribute, added by LocalRootSpanMiddleware.
    $urlScheme = $this->getAttributeValue(
      TraceAttributes::URL_SCHEME,
      $spans[0]->getAttributes(),
    );
    $this->assertNotNull($urlScheme);
    $this->assertTrue($urlScheme->hasStringValue());
    $this->assertEquals('http', $urlScheme->getStringValue());

    // Check for route span attribute, added by LocalRootSpanSubscriber.
    $httpStatus = $this->getAttributeValue(
      TraceAttributes::HTTP_ROUTE,
      $spans[0]->getAttributes()
    );
    $this->assertNotNull($httpStatus);
    $this->assertTrue($httpStatus->hasStringValue());
    $this->assertEquals('/otlrs_test/ok/{param}/edit', $httpStatus->getStringValue());

    // Check for http response status code, added by LocalRootSpanMiddleware.
    $httpStatus = $this->getAttributeValue(
      TraceAttributes::HTTP_RESPONSE_STATUS_CODE,
      $spans[0]->getAttributes()
    );
    $this->assertNotNull($httpStatus);
    $this->assertTrue($httpStatus->hasIntValue());
    $this->assertEquals(403, $httpStatus->getIntValue());
  }

  /**
   * Tests that a denied page request is recorded properly.
   */
  public function testNotFound(): void {
    $this->resetFileExporterResultFiles();
    $this->drupalGet('/otlrs_test/not-found');

    [$traceMessage] = $this->waitForTraces();

    $spans = $this->getTraceSpans($traceMessage);
    assert($spans[0] instanceof Span);
    $this->assertEquals('GET', $spans[0]->getName());

    // Check for query span attribute, added by LocalRootSpanMiddleware.
    $urlScheme = $this->getAttributeValue(
      TraceAttributes::URL_SCHEME,
      $spans[0]->getAttributes(),
    );
    $this->assertNotNull($urlScheme);
    $this->assertTrue($urlScheme->hasStringValue());
    $this->assertEquals('http', $urlScheme->getStringValue());

    // Check for route span attribute, added by LocalRootSpanSubscriber.
    $httpStatus = $this->getAttributeValue(
      TraceAttributes::HTTP_ROUTE,
      $spans[0]->getAttributes()
    );
    $this->assertNull($httpStatus);

    // Check for http response status code, added by LocalRootSpanMiddleware.
    $httpStatus = $this->getAttributeValue(
      TraceAttributes::HTTP_RESPONSE_STATUS_CODE,
      $spans[0]->getAttributes()
    );
    $this->assertNotNull($httpStatus);
    $this->assertTrue($httpStatus->hasIntValue());
    $this->assertEquals(404, $httpStatus->getIntValue());
  }

}
