<?php

declare(strict_types=1);

namespace Drupal\Tests\otlrs\Unit;

use Drupal\otlrs\StackMiddleware\LocalRootSpanMiddleware;
use Drupal\Tests\otlrs\IntrospectableSpan;
use Drupal\Tests\otlrs\Traits\SpanAssertionsTrait;
use Drupal\Tests\UnitTestCase;
use PHPUnit\Framework\Attributes\After;
use PHPUnit\Framework\Attributes\Before;
use PHPUnit\Framework\Attributes\CoversClass;
use PHPUnit\Framework\Attributes\Group;
use PHPUnit\Framework\MockObject\MockObject;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\HttpKernelInterface;

/**
 * Tests OpenTelemetry log bridge.
 */
#[Group('otlrs')]
#[CoversClass(LocalRootSpanMiddleware::class)]
final class LocalRootSpanMiddlewareTest extends UnitTestCase {

  use SpanAssertionsTrait;

  /**
   * Unset environment variables which potentially influence the test result.
   */
  #[Before]
  #[After]
  public function cleanEnv(): void {
    putenv('OTEL_INSTRUMENTATION_HTTP_KNOWN_METHODS');
    putenv('OTEL_INSTRUMENTATION_HTTP_CAPTURE_HEADERS_SERVER_REQUEST');
    putenv('OTEL_INSTRUMENTATION_HTTP_CAPTURE_HEADERS_SERVER_RESPONSE');
  }

  /**
   * Test that http.request.method method is recorded.
   */
  public function testRequestMethod(): void {
    $request = Request::create('/');
    $response = new Response();

    $span = new IntrospectableSpan();
    $middleware = new LocalRootSpanMiddleware($this->createHttpKernel($response), $span);
    $middleware->handle($request, HttpKernelInterface::MAIN_REQUEST);

    $this->assertEquals('GET', $span->getName());
    $this->assertSpanContainsAttributePairs($span, [
      'http.request.method' => 'GET',
    ]);
    $this->assertSpanNotContainsAttributeKeys($span, [
      'http.request.method_original',
    ]);
  }

  /**
   * Test that request method is set to _OTHER for exotic http method.
   */
  public function testRequestMethodUnknown(): void {
    $request = Request::create('/', 'PROPFIND');
    $response = new Response();

    $span = new IntrospectableSpan();
    $middleware = new LocalRootSpanMiddleware($this->createHttpKernel($response), $span);
    $middleware->handle($request, HttpKernelInterface::MAIN_REQUEST);

    $this->assertEquals('HTTP', $span->getName());
    $this->assertSpanContainsAttributePairs($span, [
      'http.request.method' => '_OTHER',
      'http.request.method_original' => 'PROPFIND',
    ]);
  }

  /**
   * Test that list of known request methods can be customized.
   */
  public function testRequestMethodCustomized(): void {
    putenv('OTEL_INSTRUMENTATION_HTTP_KNOWN_METHODS=UNLOCK,PROPFIND,PROPPATCH');

    $request = Request::create('/', 'PROPFIND');
    $response = new Response();

    $span = new IntrospectableSpan();
    $middleware = new LocalRootSpanMiddleware($this->createHttpKernel($response), $span);
    $middleware->handle($request, HttpKernelInterface::MAIN_REQUEST);

    $this->assertEquals('PROPFIND', $span->getName());
    $this->assertSpanContainsAttributePairs($span, [
      'http.request.method' => 'PROPFIND',
    ]);
    $this->assertSpanNotContainsAttributeKeys($span, [
      'http.request.method_original',
    ]);
  }

  /**
   * Test that nothing happens for subrequests.
   */
  public function testSubrequest(): void {
    $request = Request::create('/');
    $response = new Response();

    $span = new IntrospectableSpan();
    $middleware = new LocalRootSpanMiddleware($this->createHttpKernel($response), $span);
    $middleware->handle($request, HttpKernelInterface::SUB_REQUEST);

    $this->assertEquals('', $span->getName());
    $this->assertEquals([], $span->getAttributes());
  }

  /**
   * Test that remaining request and response attributes are recorder.
   *
   * Remaining required request and response attributes include: `url.path`,
   * `url.scheme`, `url.query` and `http.response.status_code`.
   */
  public function testRemainingRequiredAttributes(): void {
    $request = Request::create('/some/actual/path', parameters: ['q' => 42]);
    $response = new Response("I'm a teapot", 418);

    $span = new IntrospectableSpan();
    $middleware = new LocalRootSpanMiddleware($this->createHttpKernel($response), $span);
    $middleware->handle($request, HttpKernelInterface::MAIN_REQUEST);

    $this->assertEquals('GET', $span->getName());
    $this->assertSpanContainsAttributePairs($span, [
      'url.path' => '/some/actual/path',
      'url.scheme' => 'http',
      'url.query' => 'q=42',
      'http.response.status_code' => 418,
    ]);
  }

  /**
   * Test that recommended attributes are recorded by default.
   */
  public function testRecommendedAttributes(): void {

    $savedTrustedProxies = Request::getTrustedProxies();
    $savedTrustedHeaders = Request::getTrustedHeaderSet();
    try {
      Request::setTrustedProxies(['127.0.0.2', '127.0.0.3'], Request::HEADER_FORWARDED);
      $request = Request::create('/', 'POST', server: [
        'HTTP_FORWARDED' => 'for=203.0.113.1;proto=https',
        'HTTP_HOST' => 'example.com',
        'HTTP_USER_AGENT' => 'curl/7.54.1',
        'HTTP_VIA' => 'HTTP/2.0 vegur',
        'REMOTE_ADDR' => '127.0.0.2',
        'REMOTE_PORT' => '12345',
        'SERVER_PROTOCOL' => 'HTTP/1.1',
      ]);
      $response = new Response();

      $span = new IntrospectableSpan();
      $middleware = new LocalRootSpanMiddleware($this->createHttpKernel($response), $span);
      $middleware->handle($request, HttpKernelInterface::MAIN_REQUEST);

      $this->assertSpanContainsAttributePairs($span, [
        'client.address' => '203.0.113.1',
        'network.peer.address' => '127.0.0.2',
        'network.protocol.version' => '2.0',
        'server.address' => 'example.com',
        'server.port' => '443',
        'user_agent.original' => 'curl/7.54.1',
      ]);
    }
    finally {
      // Restore trusted proxies.
      Request::setTrustedProxies($savedTrustedProxies, $savedTrustedHeaders);
    }
  }

  /**
   * Test that no opt-in request and response headers are recorded by default.
   */
  public function testNoOptInHeaderAttributesByDefault(): void {
    $request = Request::create('/', 'GET', server: [
      'HTTP_REFERER' => 'http://example.com/checkout',
      'HTTP_X_REQUESTED_WITH' => 'XMLHttpRequest',
    ]);
    $response = new Response('{}', headers: [
      'HTTP_CONTENT_TYPE' => 'application/json',
    ]);
    $response->setPublic()->setMaxAge(600);

    $span = new IntrospectableSpan();
    $middleware = new LocalRootSpanMiddleware($this->createHttpKernel($response), $span);
    $middleware->handle($request, HttpKernelInterface::MAIN_REQUEST);

    foreach (array_keys($span->getAttributes()) as $key) {
      $this->assertStringStartsNotWith('http.request.header.', $key);
      $this->assertStringStartsNotWith('http.response.header.', $key);
    }
  }

  /**
   * Test that no opt-in request and response headers are recorded by default.
   */
  public function testOptInHeaderAttributes(): void {
    putenv('OTEL_INSTRUMENTATION_HTTP_CAPTURE_HEADERS_SERVER_REQUEST=X-Requested-With,User-Agent');
    putenv('OTEL_INSTRUMENTATION_HTTP_CAPTURE_HEADERS_SERVER_RESPONSE=Content-Type,Content-Encoding,X-Drupal-Cache');

    $request = Request::create('/', 'GET', server: [
      'HTTP_REFERER' => 'http://example.com/checkout',
      'HTTP_X_REQUESTED_WITH' => 'XMLHttpRequest',
    ]);
    $response = new JsonResponse(headers: [
      'X-Drupal-Cache' => 'HIT',
    ]);
    $response->setPublic()->setMaxAge(600);

    $span = new IntrospectableSpan();
    $middleware = new LocalRootSpanMiddleware($this->createHttpKernel($response), $span);
    $middleware->handle($request, HttpKernelInterface::MAIN_REQUEST);

    $this->assertSpanContainsAttributePairs($span, [
      'http.request.header.x-requested-with' => ['XMLHttpRequest'],
      'http.response.header.content-type' => ['application/json'],
      'http.response.header.x-drupal-cache' => ['HIT'],
    ]);
    $this->assertSpanNotContainsAttributeKeys($span, [
      'http.request.header.referer',
      'http.response.header.cache-control',
    ]);
  }

  /**
   * Creates a http kernel mock object.
   */
  protected function createHttpKernel(Response $response): HttpKernelInterface|MockObject {
    $kernel = $this->createMock(HttpKernelInterface::class);
    $kernel->expects($this->once())
      ->method('handle')
      ->willReturn($response);
    return $kernel;
  }

}
