<?php

declare(strict_types=1);

namespace Drupal\Tests\otlrs\Unit;

use Drupal\Core\Routing\RouteObjectInterface;
use Drupal\otlrs\EventSubscriber\LocalRootSpanSubscriber;
use Drupal\Tests\otlrs\IntrospectableSpan;
use Drupal\Tests\UnitTestCase;
use PHPUnit\Framework\Attributes\CoversClass;
use PHPUnit\Framework\Attributes\Group;
use PHPUnit\Framework\MockObject\MockObject;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Event\ExceptionEvent;
use Symfony\Component\HttpKernel\Event\RequestEvent;
use Symfony\Component\HttpKernel\HttpKernelInterface;
use Symfony\Component\Routing\Route;

/**
 * Tests OpenTelemetry log bridge.
 */
#[Group('otlrs')]
#[CoversClass(LocalRootSpanSubscriber::class)]
final class LocalRootSpanSubscriberTest extends UnitTestCase {

  /**
   * Test that nothing happens if request has no route.
   */
  public function testRequestWithoutRoute(): void {
    $request = Request::create('/');

    $span = new IntrospectableSpan();
    $subscriber = new LocalRootSpanSubscriber($span);
    $event = new RequestEvent($this->createHttpKernel(), $request, HttpKernelInterface::MAIN_REQUEST);
    $subscriber->onKernelRequest($event);

    $this->assertEquals('', $span->getName());
    $this->assertEquals([], $span->getAttributes());
    $this->assertNull($span->getException());
  }

  /**
   * Test that attribtes are updated if request has a route.
   */
  public function testRequestWithRoute(): void {
    $route = new Route('/some/{template}/path');
    $request = Request::create('/some/actual/path');

    $request->attributes->set(RouteObjectInterface::ROUTE_OBJECT, $route);
    $request->attributes->set(RouteObjectInterface::ROUTE_NAME, 'some.path');

    $span = new IntrospectableSpan();
    $subscriber = new LocalRootSpanSubscriber($span);
    $event = new RequestEvent($this->createHttpKernel(), $request, HttpKernelInterface::MAIN_REQUEST);
    $subscriber->onKernelRequest($event);

    $this->assertEquals('GET /some/{template}/path', $span->getName());
    $this->assertEquals(['http.route' => '/some/{template}/path'], $span->getAttributes());
    $this->assertNull($span->getException());
  }

  /**
   * Test that nothing happens for subrequests.
   */
  public function testSubrequestWithRoute(): void {
    $route = new Route('/some/{template}/path');
    $request = Request::create('/some/actual/path');

    $request->attributes->set(RouteObjectInterface::ROUTE_OBJECT, $route);
    $request->attributes->set(RouteObjectInterface::ROUTE_NAME, 'some.path');

    $span = new IntrospectableSpan();
    $subscriber = new LocalRootSpanSubscriber($span);
    $event = new RequestEvent($this->createHttpKernel(), $request, HttpKernelInterface::SUB_REQUEST);
    $subscriber->onKernelRequest($event);

    $this->assertEquals('', $span->getName());
    $this->assertEquals([], $span->getAttributes());
    $this->assertNull($span->getException());
  }

  /**
   * Test that an exception is recorded even if request has no route.
   */
  public function testExceptionWithoutRoute(): void {
    $request = Request::create('/');

    $exception = new \RuntimeException('Route not found.');

    $span = new IntrospectableSpan();
    $subscriber = new LocalRootSpanSubscriber($span);
    $event = new ExceptionEvent($this->createHttpKernel(), $request, HttpKernelInterface::MAIN_REQUEST, $exception);
    $subscriber->onKernelException($event);

    $this->assertEquals('', $span->getName());
    $this->assertEquals([], $span->getAttributes());
    $this->assertEquals($exception, $span->getException());
  }

  /**
   * Test that attribtes are updated and exception is recorded.
   */
  public function testExceptionWithRoute(): void {
    $route = new Route('/some/{template}/path');
    $request = Request::create('/some/actual/path');

    $request->attributes->set(RouteObjectInterface::ROUTE_OBJECT, $route);
    $request->attributes->set(RouteObjectInterface::ROUTE_NAME, 'some.path');

    $exception = new \RuntimeException('View not found.');

    $span = new IntrospectableSpan();
    $subscriber = new LocalRootSpanSubscriber($span);
    $event = new ExceptionEvent($this->createHttpKernel(), $request, HttpKernelInterface::MAIN_REQUEST, $exception);
    $subscriber->onKernelException($event);

    $this->assertEquals('GET /some/{template}/path', $span->getName());
    $this->assertEquals(['http.route' => '/some/{template}/path'], $span->getAttributes());
    $this->assertEquals($exception, $span->getException());
  }

  /**
   * Test that nothing happens for exceptions in subrequests.
   */
  public function testExceptionInSubrequestWithRoute(): void {
    $route = new Route('/some/{template}/path');
    $request = Request::create('/some/actual/path');

    $request->attributes->set(RouteObjectInterface::ROUTE_OBJECT, $route);
    $request->attributes->set(RouteObjectInterface::ROUTE_NAME, 'some.path');

    $exception = new \RuntimeException('Route not found.');

    $span = new IntrospectableSpan();
    $subscriber = new LocalRootSpanSubscriber($span);
    $event = new ExceptionEvent($this->createHttpKernel(), $request, HttpKernelInterface::SUB_REQUEST, $exception);
    $subscriber->onKernelException($event);

    $this->assertEquals('', $span->getName());
    $this->assertEquals([], $span->getAttributes());
    $this->assertNull($span->getException());
  }

  /**
   * Creates a http kernel mock object.
   */
  protected function createHttpKernel(): HttpKernelInterface|MockObject {
    return $this->createMock(HttpKernelInterface::class);
  }

}
